/***************************************************************************
 *   Copyright (C) 2005 by Claudio Attaccalite                             *
 *   claudio@freescience.info                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef SVector_H
#define SVector_H

#include <iostream>
#include "lapack_wrapper.h"
#include "blas_wrapper.h"

using namespace std;


/**
  A 1-dimensional template for making arrays.  All of the memory allocation
  and deallocation details are dealt with by the class.
*/

class SVector
{
 public:
	  friend class SMatrix;

 private:
  /**
     Number of elements in the array's first dimension.
  */

  int n_1;


  /**
     Array containing the data.
  */

  float* pArray;


  /** 
   * leading dimension is always 1 in this simple class
   **/
  
  int lda;

public:
  /**
    Gets the number of elements in the array's first dimension.

    @return number of elements in the array's first dimension.
    */
  int dim1(){return n_1;}

  /**
    Gets the total number of elements in the array.

    @return total number of elements in the array.
    */
  int size(){return n_1;}

  /**
    Gets a pointer to an array containing the array elements.  floathe ordering of this
    array is NOfloat specified.  
    */
  float* array(){return pArray;}

  /**
     Allocates memory for the array.

     @param i size of the array's first dimension.
  */

  void allocate(int i)
    {
      if( n_1 != i )
	{
	  deallocate();
	  
	  n_1 = i;
	  
	  if(n_1 >= 1) 
	    {
	      pArray = new float[n_1];
	    }
	  else 
	    {
	      n_1 = 0;
	      pArray = 0;
	    }
	}
      lda=1;
    }


  /**
     Deallocates memory for the array.
  */

  void deallocate()
    {
      if( n_1 > 0 )
	{
	  delete [] pArray;
	  pArray = 0;

	  n_1 = 0;
	}
    }

   
   /**
    * Returns the Euclidean Norm of the vector
    **/

    float Norm()
    {
      return F77_FUNC(snrm2,SNRM2)(&n_1,pArray,&lda);
    }
 
   /**
    * Returns the Absolute Sum of the vector
    **/

    float AbsoluteSum()
    {
      return F77_FUNC(sasum,SASUM)(&n_1,pArray,&lda);
    }
   
    /**
    * Returns the Absolute Sum of the vector
    **/

    float Sum()
    {
      float sum=0;
      for(int i=0;i<n_1;i++) sum+=pArray[i];
      return sum;
    }


   /**
    * Returns the index of the largest element in pArray
    **/

    int iMax()
    {
      return F77_FUNC(isamax,ISAMAX)(&n_1,pArray,&lda);
    }


  /**
     Sets two arrays equal.
  */

  void operator=(const SVector & rhs)
  {
   if(n_1 != rhs.n_1) allocate(rhs.n_1);
   F77_FUNC(scopy,SCOPY)(&n_1,rhs.pArray,&lda,pArray,&lda);
  }
  
  /**
     Sets all of the elements in an array equal to the same value.
  */

  void operator=(const float C)
  {
    for(int i=0; i<n_1; i++)
      {
	pArray[i] = C;
      }
  }
  

  /**
     Returns the dot product of two arrays.
  */

  float operator*( const SVector & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 )
	{
	  cerr << "ERROR: incorrect sizes for two SVector's in SVector*"
	       << "SVector" << endl;
	  exit(1);
	}
      if( n_1 <= 0 )
	{
	  cerr << "ERROR: SVector of size 0 used in SVector*SVector" << endl;
	  exit(1);
	}
#endif
      return  F77_FUNC(sdot,SDOT)(&n_1,pArray,&lda,rhs.pArray,&lda);
    }


  /**
     Returns the product of an array and a float.
  */

  SVector operator*(float rhs)
    {
      SVector  A(n_1);
      float zero=0.0;
      F77_FUNC(sscal,SSCAL)(&n_1,&zero,A.pArray,&lda);
      F77_FUNC(saxpy,SAXPY)(&n_1,&rhs,pArray,&lda,A.pArray,&lda);
      return A;
    }


  /**
     Returns the sum of two arrays.
  */

  SVector operator+( const SVector & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 )
	{
	  cerr << "ERROR: incorrect sizes for two SVector's in SVector+"
	       << "SVector" << endl;
	  exit(1);
	}
      if( n_1 <= 0 )
	{
	  cerr << "ERROR: SVector of size 0 used in SVector+SVector" << endl;
	  exit(1);
	}
#endif
      SVector  A(n_1);
      float d_one=1.0;
      F77_FUNC(scopy,SCOPY)(&n_1,pArray,&lda,A.pArray,&lda);
      F77_FUNC(saxpy,SAXPY)(&n_1,&d_one,rhs.pArray,&lda,A.pArray,&lda);
      return A;
    }


  /**
     Returns the difference of two arrays.
  */

  SVector operator-( const SVector & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 )
	{
	  cerr << "ERROR: incorrect sizes for two SVector's in SVector-"
	       << "SVector" << endl;
	  exit(1);
	}
      if( n_1 <= 0 )
	{
	  cerr << "ERROR: SVector of size 0 used in SVector-SVector" << endl;
	  exit(1);
	}
#endif
      SVector  A(n_1);
      float d_m_one=-1.0;
      F77_FUNC(scopy,SCOPY)(&n_1,pArray,&lda,A.pArray,&lda);
      F77_FUNC(saxpy,SAXPY)(&n_1,&d_m_one,rhs.pArray,&lda,A.pArray,&lda);
      return A;
    }


  /**
     Sets this array equal to itself times a scalar value.
  */

  void operator*=(float C)
    {
    F77_FUNC(sscal,SSCAL)(&n_1,&C,pArray,&lda);
    }


  /**
     Sets this array equal to itself divided by a scalar value.
  */

  void operator/=(float C)
    {
    float inv=1.0/C;
    F77_FUNC(sscal,SSCAL)(&n_1,&inv,pArray,&lda);
    }


  /**
     Creates an array.
  */

  SVector(){pArray = 0; n_1 = 0; lda=1;}


  /**
     Creates an array and allocates memory.
     
     @param i size of the array's first dimension.
  */

  SVector(int i){pArray = 0; n_1 = 0; allocate(i); lda=1; }


  /**
     Creates an array and sets it equal to another array.

     @param rhs array to set this array equal to.
  */

  SVector(const SVector & rhs)
    {
      n_1 = 0;
      pArray = 0;
      lda = 1;
      allocate(rhs.n_1);
      F77_FUNC(scopy,SCOPY)(&n_1,rhs.pArray,&lda,pArray,&lda);
    }

  /**
     Creates an array and sets it equal to a C array.

     @param C array to set this array equal to.
  */

  SVector(int dim,float *rhs)
    {
      n_1 = 0;
      pArray = 0;
      lda = 1;
      allocate(dim);
      n_1=dim;
      F77_FUNC(scopy,SCOPY)(&n_1,rhs,&lda,pArray,&lda);
    }

   
  /**
     Copy all elements from a C array in the Vector
  */

  void set(int ndim,float *rhs)
    {
#ifdef DEBUG
       if( n_1 != ndim )
	{
	  cerr << "ERROR: incorrect sizes for two Vector's in "
               << " SVector.set(n_1,c_array) " << endl;
	  exit(1);
	}
#endif
      F77_FUNC(scopy,SCOPY)(&n_1,rhs,&lda,pArray,&lda);
    }
  
  /**
     Copy all elements in a C array from the Vector
  **/

  void get(int dim,float *rhs)
    {
#ifdef DEBUG
       if( n_1 != dim )
	{
	  cerr << "ERROR: incorrect sizes for two Vector's in "
               << " Vector.get(n_1,c_array) " << endl;
	  exit(1);
	}
#endif
      F77_FUNC(scopy,SCOPY)(&n_1,pArray,&lda,rhs,&lda);
    }
  
  /**
     Destroy's the array and cleans up the memory.
  */

  ~SVector(){deallocate();}

  /**
    Accesses element <code>(i)</code> of the array.
    */
  float& operator()(int i){return pArray[i];}

  /**
     Prints the array to a stream.
  */

  friend ostream& operator<<(ostream & strm, const SVector & rhs)
    {
      for( int i=0; i<rhs.n_1; i++ )  
	{
	  strm << rhs.pArray[i] << "\t";
	}
      strm << endl;
      return strm;
    }
};

#endif






