/***************************************************************************
 *   Copyright (C) 2004 by Claudio Attaccalite                             *
 *   claudio@freescience.info                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifndef ZMatrix_H
#define ZMatrix_H
#if HAVE_CONFIG
#include <config.h>
#endif

#include <iostream>
#include "ZVector.h"
#include "lapack_wrapper.h"
#include "blas_wrapper.h"


using namespace std;


/**
  A 2-dimensional template for making arrays.  All of the memory allocation
  and deallocation details are dealt with by the class. Fortran allocation
*/


class ZMatrix
{
  private:
  /**
     Number of elements in the array's first dimension.
  */

  int n_1;


  /**
     Number of elements in the array's second dimension.
  */

  int n_2;

  /**
     Array containing the data.
  */

  complex<double>* pArray;

  /**
    Temporaney pointer to swap coloumns
  */
  
  complex<double>* tmpArray;

  /**
   Temporaney variable to transpose matrix
    */
  complex<double> temp;

  /**
   * one dimension, in this simple class it is always one
   **/
  int one;

public:
  /**
    Gets the number of elements in the array's first dimension.

    @return number of elements in the array's first dimension.
    */
  int dim1(){return n_1;}

  /**
    Gets the number of elements in the array's second dimension.

    @return number of elements in the array's second dimension.
    */
  int dim2(){return n_2;}

  /**
    Gets the total number of elements in the array.

    @return total number of elements in the array.
    */
  int size(){return n_1*n_2;}

  /**
    Gets a pointer to a col the array elements.  The ordering of this
    array is NOT specified.  
    */
  complex<double>* array(){return pArray;}

  /**
    Transpose the 2D matrix
  */ 
  
 void transpose()
 {
 if(n_1 !=n_2)
 {
	 cerr << " Transpose only for square matrix ! " << endl;
	 exit(1);
 }
 for(int i=0;i<n_1;i++)
	for(int j=0;j<i;j++)
	{
		temp=pArray[j+i*n_1];
		pArray[j+i*n_1]=pArray[i+j*n_1];
		pArray[i+j*n_1]=temp;
	}
 }

  /**
     Allocates memory for the array.

     @param i size of the array's first dimension.
     @param j size of the array's second dimension.
  */
  
  void allocate(int i, int j)
    {
      if( n_1 != i || n_2 != j )
	{
	  deallocate();
	  
	  n_1 = i;
	  n_2 = j;
	  
	  if(n_1 >= 1 && n_2 >= 1)
	      pArray = new complex<double> [n_1*n_2];
	  else 
	      pArray = 0;
	}
    }

  /**
     Deallocates memory for the array.
  */

  void deallocate()
    {
      if(n_2>0 && n_1>0)
      {
       delete [] pArray;
      }
       pArray = 0;

      n_1 = 0;
      n_2 = 0;
    }


  /**
     Sets two matrix equal.
  */

   void operator=(const ZMatrix & rhs)
    {
     int tot=n_1*n_2;
#ifdef DEBUG
     if(n_1 != rhs.n_1 || n_2 != rhs.n_2) allocate(rhs.n_1,rhs.n_2);
#endif
    
    F77_FUNC(zcopy,ZCOPY)(&tot,(double*)rhs.pArray,&one,(double*)pArray,&one);
    }


  /**
     Sets all of the elements in an array equal to the complex value.
  */

  void operator=(const complex<double> C)
    {
      for(int i=0; i<n_1*n_2;i++) pArray[i] = C;
    }
  
 /**
     Sets all of the elements in an array equal to the real value.
  */

  void operator=(const double C)
    {
      for(int i=0; i<n_1*n_2;i++) pArray[i] = dcmplx(C,0);
    }

  /**
     Scale all elements for C
  */
  

  void operator*=(complex<double> C)
    {
     	int tot=n_1*n_2;
	F77_FUNC(zscal,ZSCAL)(&tot,(double*)&C,(double*)pArray,&one);
    }

  /**
     Sets this array equal to itself divided by a scalar value.
  */

  void operator/=(const complex<double> C)
    {
     	int tot=n_1*n_2;
	complex<double> inv=dcmplx(1.0,0.0)/C;
	F77_FUNC(zscal,ZSCAL)(&tot,(double*)&inv,(double*)pArray,&one);
    }


  /**
     Creates an array.
  */

  ZMatrix(){pArray = 0; n_1 = 0; n_2 = 0; one = 1;}


  /**
     Creates an array and allocates memory.
     
     @param i size of the array's first dimension.
     @param j size of the array's second dimension.
  */

  ZMatrix(int i, int j) {pArray = 0; n_1 = 0; n_2 = 0; one=1; allocate(i,j);}


  /**
     Creates an array and sets it equal to another array.

     @param rhs array to set this array equal to.
  */

  ZMatrix( const ZMatrix & rhs)
    {
      n_1 = 0;
      n_2 = 0;
      pArray = 0;
      allocate(rhs.n_1,rhs.n_2);
      int tot=n_1*n_2;
      one=1;
      F77_FUNC(zcopy,ZCOPY)(&tot,(double*)rhs.pArray,&one,(double*)pArray,&one);
    }


  /**
     Destroy's the array and cleans up the memory.
  */

  ~ZMatrix(){deallocate();}

 /**
  * Sum all elements of the matrix
  **/
  complex<double> SumAll()
  {
   complex<double> sum=0;
   for(int i=0;i<n_1*n_2;i++) sum+=pArray[i];
   return sum;
  }
 
 /**
  * Return the ZMatrix Trace
  **/
  complex<double> Trace()
  {
#ifdef DEBUG
   if(n_1 !=n_2)
   {
	 cerr << " Trace only for Squared ZMatrix ! " << endl;
	 exit(1);
   } 
#endif
   complex<double> sum=0;
   for(int i=0;i<n_1;i++) sum+=pArray[i*n_1+i];
   return sum;
  }
  
 /**
  * Copy an 1D array into the i-th row
  **/
  void setrow(int irow,ZVector &vector)
  {
#ifdef DEBUG
	  if(vector.n_1!=n_2 || irow>=n_1)
	  { 
		  cerr << " ERROR! dimension array copying row! " << endl;
   		  exit(1);
	  }
#endif
   F77_FUNC(zcopy,ZCOPY)(&n_2,(double*)vector.pArray,&one,(double*)(pArray+irow),&n_1);
  }

 /**
  * Copy an 1D double array into the i-th row
  **/
  /**
  void setrow(int irow,complex<double> *vector)
  {
   F77_FUNC(dcopy,DCOPY)(&n_2,vector,&one,pArray+irow,&n_1);
  }

 /**
  * Copy an 1D C double array into the i-th row with lda
  **/
  /**
  void setrow(int irow,complex<double> *vector,int lda)
  {
   F77_FUNC(dcopy,DCOPY)(&n_2,vector,&lda,pArray+irow,&n_1);
  }
  
  /**
  * Copy an 1D array into the i-th coloumn
  **/
  void setcol(int icol, const ZVector & vector)
  {
#ifdef DEBUG
	  if(vector.n_1!=n_1 || icol>=n_2)
	  { 
		  cerr << " ERROR! dimension array copying coloumn! " << endl;
   		  exit(1);
	  }
#endif
   F77_FUNC(zcopy,ZCOPY)(&n_1,(double*)vector.pArray,&one,(double*)(pArray+icol*n_1),&one);
  }

  /**
  * Copy an 1D array into the i-th coloumn
  **/
  /**
  void setcol(int icol,complex<double> *vector)
  {
#ifdef DEBUG
	  if(icol>=n_2)
	  { 
		  cerr << " ERROR! dimension array copying coloumn! " << endl;
   		  exit(1);
	  }
#endif
   F77_FUNC(dcopy,DCOPY)(&n_1,vector,&one,pArray+icol*n_1,&one);
  }
 
 /**
  * Rescale a row
  **/
  void row_scale(int irow, complex<double> C)
  {
#ifdef DEBUG
	  if(irow>=n_1)
	  { 
		  cerr << " ERROR! dimension array in row_scale! " << endl;
   		  exit(1);
	  }
#endif
   F77_FUNC(zscal,ZSCAL)(&n_2,(double*)&C,(double*)(pArray+irow),&n_1);
  }



  
 /**
  * Dot product between the i-th row and a vector
  **/
  complex<double> rowdotu(int irow, const ZVector & vector)
  {
#ifdef DEBUG
	  if(vector.n_1!=n_2 || irow>=n_1)
	  { 
		  cerr << " ERROR! dimension array in row dot! " << endl;
   		  exit(1);
	  }
#endif
   return F77_FUNC(zdotu,ZDOTU)(&n_2,(double*)(pArray+irow),&n_1,(double*)vector.pArray,&one);
  }

 /**
  * Dot product between the i-th row and a vector
  **/
  complex<double> rowdotc(int irow, const ZVector & vector)
  {
#ifdef DEBUG
	  if(vector.n_1!=n_2 || irow>=n_1)
	  { 
		  cerr << " ERROR! dimension array in row dot! " << endl;
   		  exit(1);
	  }
#endif
   return F77_FUNC(zdotc,ZDOTC)(&n_2,(double*)(pArray+irow),&n_1,(double*)vector.pArray,&one);
  }

  
  /**
  * Dot product between the i-th col and a vector
  **/
  complex<double> coldotu(int icol, const ZVector &vector)
  {
#ifdef DEBUG
	  if(vector.n_1!=n_1 || icol>=n_2)
	  { 
		  cerr << " ERROR! dimension array in coloumn dot! " << endl;
   		  exit(1);
	  }
#endif
   return F77_FUNC(zdotu,ZDOTU)(&n_1,(double*)(pArray+icol*n_1),&one,(double*)vector.pArray,&one);
  }
 
  
  /**
  * Dot product between the i-th col and a vector
  **/
  complex<double> coldotc(int icol, const ZVector &vector)
  {
#ifdef DEBUG
	  if(vector.n_1!=n_1 || icol>=n_2)
	  { 
		  cerr << " ERROR! dimension array in coloumn dot! " << endl;
   		  exit(1);
	  }
#endif
   return F77_FUNC(zdotc,ZDOTC)(&n_1,(double*)(pArray+icol*n_1),&one,(double*)vector.pArray,&one);
  }
  
  /**
  * Rescale a column
  **/
  void col_scale(int icol, complex<double> C)
  {
#ifdef DEBUG
	  if(icol>=n_2)
	  { 
		  cerr << " ERROR! dimension array in col_rescale! " << endl;
   		  exit(1);
	  }
#endif
   F77_FUNC(zscal,ZSCAL)(&n_1,(double*)&C,(double*)(pArray+icol*n_1),&one);
  }


 /**
  * Dot product between the i-th row and a vector
  **/
  /**
  complex<double> rowdot(int irow, complex<double> *vector)
  {
   return F77_FUNC(zdot,ZDOT)(&n_2,pArray+irow,&n_1,vector,&one);
  }

  /**
  * Dot product between the i-th col and a vector
  **/
  /**
  complex<double> coldot(int icol, complex<double> *vector)
  {
   return F77_FUNC(ddot,DDOT)(&n_1,pArray+icol*n_1,&one,vector,&one);
  }



/**
  * Return a pointer to the row irow
  **/
  complex<double> *row_pointer(int irow)
  {
#ifdef DEBUG
	  if(irow>=n_1)
	  { 
		  cerr << " ERROR! dimension array getting a row! " << endl;
   		  exit(1);
	  }
#endif
   return(pArray+irow);
  }
  
 /**
  * Get a copy 1D array of the i-th row
  **/
  ZVector row(int irow)
  {
#ifdef DEBUG
	  if(irow>=n_1)
	  { 
		  cerr << " ERROR! dimension array getting a row! " << endl;
   		  exit(1);
	  }
#endif
   ZVector TEMP(n_2);
   F77_FUNC(zcopy,ZCOPY)(&n_2,(double*)(pArray+irow),&n_1,(double*)TEMP.pArray,&one);
   return TEMP;
  }




 /**
  * Return a pointer to the coloumn icol
  **/
  complex<double> *col_pointer(int icol)
  {
#ifdef DEBUG
	  if(icol>=n_2)
	  { 
		  cerr << " ERROR! dimension array getting a coloumn! " << endl;
   		  exit(1);
	  }
#endif
   return(pArray+icol*n_1);
  }


  
  /**
  * Get a copy 1D array of the i-th coloumn
  **/
  ZVector col(int icol)
  {
#ifdef DEBUG
	  if(icol>=n_2)
	  { 
		  cerr << " ERROR! dimension array getting a coloumn! " << endl;
   		  exit(1);
	  }
#endif
   ZVector TEMP(n_1);
   F77_FUNC(zcopy,ZCOPY)(&n_1,(double*)(pArray+icol*n_1),&one,(double*)TEMP.pArray,&one);
   return TEMP;
  }

  /**
  * Sum all elements of a Row the matrix
  **/
  complex<double> SumRow(const int irow)
  {
#ifdef DEBUG
	  if(irow>=n_1)
	  { 
		  cerr << " ERROR! dimension array getting a row! " << endl;
   		  exit(1);
	  }
#endif
   complex<double> sum=0;
   for(int i=0;i<n_2;i++) sum+=pArray[irow+i*n_1];
   return sum;
  }


  /**
  * Sum all elements of a Col the matrix
  **/
  complex<double> SumCol(const int icol)
  {
#ifdef DEBUG
	  if(icol>=n_2)
	  { 
		  cerr << " ERROR! dimension array getting a row! " << endl;
   		  exit(1);
	  }
#endif
   complex<double> sum=0;
   for(int i=0;i<n_1;i++) sum+=pArray[i+icol*n_1];
   return sum;
  }





  /**
     Returns the sum of two matrices.
  */

  ZMatrix operator+( const ZMatrix & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 || n_2 !=rhs.n_2 )
	{
	  cerr << "ERROR: incorrect sizes for two ZMatrix's in ZMatrix+"
	       << "ZMatrix" << endl;
	  exit(1);
	}
      if( n_1 <= 0 || n_2 <=0 )
	{
	  cerr << "ERROR: ZMatrix of size 0 used in ZMatrix+ZMatrix" << endl;
	  exit(1);
	}
#endif
      int tot=n_1*n_2;
      complex<double> d_one=dcmplx(1.0,0.0);
      ZMatrix  A(n_1,n_2);
      F77_FUNC(zcopy,ZCOPY)(&tot,(double*)pArray,&one,(double*)A.pArray,&one);
      F77_FUNC(zaxpy,ZAXPY)(&tot,(double*)&d_one,(double*)rhs.pArray,&one,(double*)A.pArray,&one); 
      return A;
    }


  /**
     Returns the difference of two Matrices.
  */

  ZMatrix operator-( const ZMatrix & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 || n_2 !=rhs.n_2 )
	{
	  cerr << "ERROR: incorrect sizes for two ZMatrix's in ZMatrix-"
	       << "ZMatrix" << endl;
	  exit(1);
	}
      if( n_1 <= 0 || n_2 <=0)
	{
	  cerr << "ERROR: ZMatrix of size 0 used in ZMatrix-ZMatrix" << endl;
	  exit(1);
	}
#endif
      ZMatrix A(n_1,n_2);
      int tot=n_1*n_2;
      complex<double> d_m_one=dcmplx(-1.0,0.0);
      F77_FUNC(zcopy,ZCOPY)(&tot,(double*)pArray,&one,(double*)A.pArray,&one);
      F77_FUNC(zaxpy,ZAXPY)(&tot,(double*)&d_m_one,(double*)rhs.pArray,&one,(double*)A.pArray,&one); 
      return A;
    }

  
  /**
    Accesses element <code>(i,j)</code> of the array.
    */
  
  complex<double> & operator()(int i,int j)
  {
#ifdef DEBUG
	  if(i>=n_1 || j>=n_2)
	  { 
		  cerr << " ERROR! index out of range! " 
		       << "("<<i<<","<<j<<")" <<  endl;
   		  exit(1);
	  }
#endif
	  return pArray[i+j*n_1];
  }


  /**
     Prints the array to a stream.
  */

  friend ostream& operator<<(ostream & strm, const ZMatrix & rhs)
    {
      for(int i=0; i<rhs.n_1;i++)
	{
	  for(int j=0; j<rhs.n_2; j++)
	    {
	      strm << rhs.pArray[i+j*rhs.n_1] << "\t";
	    }
	  strm << endl;
	}
      return strm;
    }

    
};

#endif
