/***************************************************************************
 *   Copyright (C) 2005 by Claudio Attaccalite                             *
 *   claudio@freescience.info                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ZVector_H
#define ZVector_H

#include <iostream>
#include "DVector.h"
#include "lapack_wrapper.h"
#include "blas_wrapper.h"

using namespace std;


/**
  A 1-dimensional template for making arrays.  All of the memory allocation
  and deallocation details are dealt with by the class.
*/
class ZVector
{
 public:
	  friend class ZMatrix;

 private:
  /**
     Number of elements in the array's first dimension.
  */

  int n_1;


  /**
     Array containing the data.
  */

  complex<double>* pArray;


  /** 
   * leading dimension is always 1 in this simple class
   **/
  
  int lda;

public:
  /**
    Gets the number of elements in the array's first dimension.

    @return number of elements in the array's first dimension.
    */
  int dim1(){return n_1;}

  /**
    Gets the total number of elements in the array.

    @return total number of elements in the array.
    */
  int size(){return n_1;}

  /**
    Gets a pointer to an array containing the array elements.  doublehe ordering of this
    array is NOdouble specified.  
    */
  complex<double>* array(){return pArray;}

  /**
     Allocates memory for the array.

     @param i size of the array's first dimension.
  */

  void allocate(int i)
    {
      if( n_1 != i )
	{
	  deallocate();
	  
	  n_1 = i;
	  
	  if(n_1 >= 1) 
	    {
	      pArray = new complex<double>[n_1];
	    }
	  else 
	    {
	      n_1 = 0;
	      pArray = 0;
	    }
	}

      lda = 1;
    }


  /**
     Deallocates memory for the array.
  */

  void deallocate()
    {
      if( n_1 > 0 )
	{
	  delete [] pArray;
	  pArray = 0;

	  n_1 = 0;
	}
    }

   
   /**
    * Returns the Euclidean Norm of the vector
    **/

    double Norm()
    {
      return F77_FUNC(dznrm2,DZNRM2)(&n_1,(double*)pArray,&lda);
    }
 
   /**
    * Returns the Absolute Sum of the vector
    **/

    double AbsoluteSum()
    {
      return F77_FUNC(dzasum,DZASUM)(&n_1,(double*)pArray,&lda);
    }
   
    /**
    * Returns the Sum of the vector
    **/

    complex<double> Sum()
    {
      complex<double> sum=cmplx(0.0,0.0);
      for(int i=0;i<n_1;i++) sum=sum+pArray[i];
      return sum;
    }
    
    /**
    * Returns the Real part of the vector
    **/

    DVector real()
    {
      DVector temp(n_1);
      for(int i=0;i<n_1;i++) temp(i)=pArray[i].real();
      return temp;
    }

   /**
    * Returns the immaginary part of the vector
    **/

    DVector imag()
    {
      DVector temp(n_1);
      for(int i=0;i<n_1;i++) temp(i)=pArray[i].imag();
      return temp;
    }


   /**
    * Returns the index of the largest element in pArray
    **/

    int iMax()
    {
      return F77_FUNC(izamax,IZAMAX)(&n_1,(double*)pArray,&lda);
    }


  /**
     Sets two arrays equal.
  */

  void operator=(const ZVector & rhs)
  {
   if(n_1 != rhs.n_1) allocate(rhs.n_1);
   F77_FUNC(zcopy,ZCOPY)(&n_1,(double*)rhs.pArray,&lda,(double*)pArray,&lda);
  }
  
  /**
     Sets all of the elements in an array equal to the same value.
  */

  void operator=(const double C)
  {
    for(int i=0; i<n_1; i++)
      {
	pArray[i] = cmplx(C,0);
      }
  }
  
   /**
     Sets all of the elements in an array equal to the same value.
  */

  void operator=(const complex<double> C)
  {
    for(int i=0; i<n_1; i++)
      {
	pArray[i] = C;
      }
  }
  
  /**
     Returns the dot product of two arrays.
  */

  complex<double> zdotu(const ZVector & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 )
	{
	  cerr << "ERROR: incorrect sizes for two ZVector's in ZVector*"
	       << "ZVector" << endl;
	  exit(1);
	}
      if( n_1 <= 0 )
	{
	  cerr << "ERROR: ZVector of size 0 used in ZVector*ZVector" << endl;
	  exit(1);
	}
#endif
      return  F77_FUNC(zdotu,ZDOTU)(&n_1,(double*)pArray,&lda,(double*)rhs.pArray,&lda);
    }

    complex<double> zdotc(const ZVector & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 )
	{
	  cerr << "ERROR: incorrect sizes for two ZVector's in ZVector*"
	       << "ZVector" << endl;
	  exit(1);
	}
      if( n_1 <= 0 )
	{
	  cerr << "ERROR: ZVector of size 0 used in ZVector*ZVector" << endl;
	  exit(1);
	}
#endif
      return  F77_FUNC(zdotc,ZDOTC)(&n_1,(double*)pArray,&lda,(double*)rhs.pArray,&lda);
    }

  /**
     Returns the sum of two arrays.
  */

  ZVector operator+( const ZVector & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 )
	{
	  cerr << "ERROR: incorrect sizes for two ZVector's in ZVector+"
	       << "ZVector" << endl;
	  exit(1);
	}
      if( n_1 <= 0 )
	{
	  cerr << "ERROR: ZVector of size 0 used in ZVector+ZVector" << endl;
	  exit(1);
	}
#endif
      ZVector  A(n_1);
      double d_one=1.0;
      F77_FUNC(zcopy,ZCOPY)(&n_1,(double*)pArray,&lda,(double*)A.pArray,&lda);
      F77_FUNC(zaxpy,ZAXPY)(&n_1,&d_one,(double*)rhs.pArray,&lda,(double*)A.pArray,&lda); 
      return A;
    }


  /**
     Returns the difference of two arrays.
  */

  ZVector operator-( const ZVector & rhs)
    {
#ifdef DEBUG
      if( n_1 != rhs.n_1 )
	{
	  cerr << "ERROR: incorrect sizes for two ZVector's in ZVector-"
	       << "ZVector" << endl;
	  exit(1);
	}
      if( n_1 <= 0 )
	{
	  cerr << "ERROR: ZVector of size 0 used in ZVector-ZVector" << endl;
	  exit(1);
	}
#endif
      ZVector  A(n_1);
      complex<double> d_m_one(-1.0,0.0);
      F77_FUNC(zcopy,ZCOPY)(&n_1,(double*)pArray,&lda,(double*)A.pArray,&lda);
      F77_FUNC(zaxpy,ZAXPY)(&n_1,(double*)&d_m_one,(double*)rhs.pArray,&lda,(double*)A.pArray,&lda); 
      return A;
    }


  /**
     Sets this array equal to itself times a scalar value.
  */

  void operator*=(complex<double> C)
    {
    F77_FUNC(zscal,ZSCAL)(&n_1,(double*)&C,(double*)pArray,&lda);
    }


  /**
     Sets this array equal to itself divided by a scalar value.
  */

  void operator/=(complex<double> C)
    {
    complex<double> inv=1.0/C;
    F77_FUNC(zscal,ZSCAL)(&n_1,(double*)&inv,(double*)pArray,&lda);
    }


  /**
     Creates an array.
  */

  ZVector(){pArray = 0; n_1 = 0; lda=1;}


  /**
     Creates an array and allocates memory.
     
     @param i size of the array's first dimension.
  */

  ZVector(int i){pArray = 0; n_1 = 0; allocate(i); lda=1; }


  /**
     Creates an array and sets it equal to another array.

     @param rhs array to set this array equal to.
  */

  ZVector(const ZVector & rhs)
    {
      n_1 = 0;
      pArray = 0;
      lda = 1;
      allocate(rhs.n_1);
      F77_FUNC(zcopy,ZCOPY)(&n_1,(double*)rhs.pArray,&lda,(double*)pArray,&lda);
    }

  /**
     Destroy's the array and cleans up the memory.
  */

  ~ZVector(){deallocate();}

  /**
    Accesses element <code>(i)</code> of the array.
    */
  complex<double>& operator()(int i){return pArray[i];}

  /**
     Prints the array to a stream.
  */

  friend ostream& operator<<(ostream & strm, const ZVector & rhs)
    {
      for( int i=0; i<rhs.n_1; i++ )  
	{
	  strm << rhs.pArray[i] << "\t";
	}
      strm << endl;
      return strm;
    }
    
};

#endif






